#!/local/gnu/bin/perl -w

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
###
### GenTree is a research product of the Information Technology
### Technology Laboratory (ITL) of the National Institute of Standards and
### Technology (NIST).  Since NIST is an agency of the Federal government,
### this software is not subject to copyright.  We ask anyone who uses or
### distributes this source code to include this notice.
###
###  Primary author: John Cugini / CUZ@NIST.GOV
###  Version Date:   2001-April-18
###
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

### ### ### ### ### ### ### ### ### ### ### ### ### ###
###
### Main purpose: Generate a td-tree.dat file for subsequent input
### to TreeDec; derive from pre-existing link structure of HTML files.
### Uses local files only, since these are the ones that TreeDec
### would process.
###
###
### Two Input files:
###    pages.dat: list of files and nicknames from convert-ll
###
###    links.dat: List of nicknames pairs, indicating linkage
###
### Output:
###    A skelton td-tree.dat file - tentative hierarchy
###
###    Messages are written to the stdout and to a message file.
###
### ### ### ### ### ### ### ### ### ### ### ### ### ###

### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub msg

# write message to file and print it out
{
    my ($p1) = @_;
    print         $p1;
    print MSGFILE $p1;
    return (0);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub errmsg

# write error message and increment error-count
{
    msg ("ERROR: $_[0] \n");
    $num_errs++;
    return (0);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub infmsg

# write info message
{
    msg ("INFO: $_[0] \n");
    return (0);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub prt_tree

# print depth-first
{
    my ($the_url, $kid, $kidstring, $lev1);
    my ($nn_parm, $level) = @_;

    if ($url_of{$nn_parm} =~ m%\A(\S+) +<\Z%)
    {  $the_url = $1;
    }
    else
    {  errmsg ("$nn_parm yields $url_of{$nn_parm} ");
       return (101);
    }

    print TREE ("\t" x $level), "$the_url   *\n";

    $kidstring = $child_of{$nn_parm};
    $lev1 = $level + 1;
    while ($kidstring =~ s%(\S+)%%)
    {  prt_tree ($1, $lev1);
    }

    return (0);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ###     MAIN-LINE   ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ###

   print "\n\n\n----------   Begin GenTree ----------\n";

   ## lame default values for parameters
   $root   = "";
   $indir  = "treedec";
   $outdir = "treedec";
   

   foreach $arg (@ARGV)
   {
      if ($arg =~ m%\Aroot=(\S+)\Z%)
      {  $root = $1;
      }
      elsif ($arg =~ m%\Aindir=(\S+)\Z%)
      {  $indir = $1;
      }
      elsif ($arg =~ m%\Aoutdir=(\S+)\Z%)
      {  $outdir = $1;
      }
      elsif ($arg =~ m%\Adir=(\S+)\Z%)
      {  $outdir = $1;
         $indir  = $1;
      }
      else
      {  print "Mystery parameter: $arg \n";
      }
   }

   $msg_name = ">$outdir/gen-td-msg.txt";
   open (MSGFILE, $msg_name) ||
      die "Could not open message file: $msg_name for output.\n";
   $num_errs = 0;

   infmsg ("Input directory = $indir");
   infmsg ("output directory = $outdir");

   if ($root eq "")
   {  errmsg ("Must specify root parameter.  I quit.");
      goto alldone;
   }
   else
   {  infmsg ("Root = $root");
   }

   open (PAGES, "<$indir/pages.dat") ||
      die "Could not open pages file for input.\n";
   open (LINKS, "<$indir/links.dat")   ||
      die "Could not open links file for input.\n";

### read in pages file
   $nn_root = "";

   ### skip host record
   $prec = <PAGES>;
   $page_ct = 0;
   while ($prec = <PAGES>)
   {
      if ($prec =~ m%\A(\S{1,12}) +([A-Z]) +(\S+)\Z%)
      {
         $nn    = $1;
         $ftype = $2;
         $url   = $3;
      }
      else
      {
         errmsg ("Cannot parse page-rec: $prec");
         next;
      }

      if ($ftype ne "H")
      {  next;
      }

      if ($url =~ m%\A/(.+)\Z%)
      {  ## must be local file
         $url = $1;
      }
      else
      {  next;
      }

      $url_of{$nn} = $url;
      $child_of{$nn} = " ";
      $page_ct++;

      if ($url eq $root)
      {  $nn_root = $nn;
      }
   }
   close (PAGES);

   if ($nn_root eq "")
   {  errmsg ("Could not find root filename: $root in pages.dat.");
      goto alldone;
   }

### read in links file
   while ($lrec = <LINKS>)
   {
      if ($lrec =~ m%\A(\S+) +(\S+)\Z%)
      {  $nn1 = $1;
         $nn2 = $2;
      }
      else
      {  errmsg ("Cannot parse link-rec: $lrec");
         next;
      }

      ## make sure both nicknames were saved from pages.dat and also
      ## that the linkage hasn't already been saved.
      if (exists $child_of{$nn1} && exists $child_of{$nn2} &&
          ($child_of{$nn1} !~ m% $nn2 %))
      {  $child_of{$nn1} .= "$nn2 ";
         ##DB print "$url_of{$nn1} --> $url_of{$nn2} \n";
         ##DB print "$nn1 --> $nn2 \n";
      }
   }
   close (LINKS);

### OK, now try to build breadth-first tree - if unconnected,
### will write error message;  build tree by knocking out
### extra child-links.

   $reached_ct = 1;
   $fifo = " $nn_root ";

   ## mark root as reached
   $url_of{$nn_root} .= " <";

   while ($fifo =~ s%\A +(\S+)%%)
   {
      $this_nn = $1;
      ##DB print "Expand from: $this_nn \n";

      $children = $child_of{$this_nn};
      $pared_kids = " ";
      while ($children =~ s%(\S+)%%)
      {
         $kid = $1;
         if ($url_of{$kid} =~ m%\A\S+ +<\Z%)
         {  ## already reached
            next;
         }

         $reached_ct++;
         ## stays as child
         $pared_kids .= "$kid ";

         ## mark as reached
         $url_of{$kid} .= " <";

         ## add to queue
         $fifo .= "$kid ";
      }

      ##DB print "pared child_of $this_nn = $pared_kids\n";
      ##DB print "Next FIFO = $fifo \n";
      $child_of{$this_nn} = $pared_kids;
   }

   if ($reached_ct == $page_ct)
   {  infmsg ("Did successful breadth-first treewalk of $page_ct pages.");
   }
   elsif ($reached_ct < $page_ct)
   {  errmsg ("Treewalk reached only $reached_ct of $page_ct pages.");
   }
   else
   {  errmsg ("Treewalk reached $reached_ct (!!) of $page_ct pages.");
   }

   ##Debug loop here

   ## now do depth-first print out.

   open (TREE, ">$outdir/td-tree.dat") ||
      die "Could not open tree file for output.\n";

   ## recursive call here
   prt_tree ($nn_root, 0);

   close (TREE);

### ### ###    we're done    ### ### ###

alldone:
   if ($num_errs == 0)
   {  infmsg ("Normal completion with no errors.");
   }
   else
   {  infmsg ("Completed with $num_errs error(s).");
   }

   close (MSGFILE);

