#!/local/gnu/bin/perl -w

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
###
###  The VISVIP system is a research product of the Information Technology
###  Laboratory of the National Institute of Standards and Technology
###  (NIST).  See http://www.nist.gov/itl/div894/vvrg/ for more details.
###  Since NIST is an agency of the Federal government, VISVIP is not
###  subject to copyright.  We ask anyone who uses or distributes this
###  source code to include this notice.
###
###  Primary author: John Cugini / CUZ@NIST.GOV
###  Version Date:   2000-nov-17
###  
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

## Parse output of Linklint (fileF.txt) to form input to VISVIP, describing
## topology of website: generate pages.dat, links.dat, and url2nn.dat.

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub subst_url
#   Arguments: current url, replacement
{
   my ($short_url, $target) = @_;

   if ($url_nn{$target})
   {
      $url_nn{$short_url}
         || die "Could not find url to be replaced: $short_url\n";
      $url_nn{$short_url} = $url_nn{$target};
      print ERRFILE "Changed nickname of $short_url to that of $target\n";
      return (0);
   }
   else
   {
      return (1);
   }  
}

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub findline
#   Argument is string to be sought in file.

{
   while ($ll_line = <LLFILE>)
   {  if ($ll_line =~ m%\A$_[0]%)
      {  return (1);
      }
   }
   return(0);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub trim_url

# knock off parameters ("?") and anchors ("#") from the end of the URL.

{
   my ($big_url) = @_;

   $big_url =~ s%\?\S*\Z%%;
   $big_url =~ s%#\S*\Z%%;
   return ($big_url);
}

### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###
sub parse_ll
#   Argument is record from LLFILE
#   return codes:
#     f - slash in cc.1 - a "link-from" address
#     t - spaces, then stuff - a "link-to" address
#     k - leading bracket - a comment line
#     c - a "contains" line
#     b - a blank line
#     0 - a "huh?" line

{
   # truncate trailing spaces, tabs, CR, FF, and newlines

   $this_line = $_[0];

   #delete newlines and then trailing spaces 
   chomp($this_line);
   $this_line =~ s%\s+\Z%%;

   if ($this_line =~ m%\A(/\S*)\Z%)
   {  $ll_url = trim_url($1);
      return ('f');
   }
   elsif ($this_line =~ m%\A\s+(\S+)\Z%) 
   {  $ll_url = trim_url($1);
      return ('t');
   }
   elsif ($this_line =~ m%\A\s+(\S+:)\s+(\S+)\Z%)
   {  $ll_url = trim_url($1 . $2);
      return ('t');
   }
   elsif ($this_line =~ m%\A\[directory listing\]\Z%)
   {  return ('d');
   }
   elsif ($this_line =~ m%\A\[file:\s+([^\]]+)\]\Z%)
   {  $ll_url = trim_url($1);
      return ('k');
   }
   elsif ($this_line =~ m%\A\s+contains ([0-9]+) links?:\Z%)
   {  return ('c');
   }
   elsif ($this_line =~ m%\A\s*\Z%)
   {  return ('b');
   }
   else
   {  return (0);
   }
}

### ### ### ### ### ### ### ### ### ### ### ### ### ###
### ### ### ### ###     MAIN-LINE   ### ### ### ### ###
### ### ### ### ### ### ### ### ### ### ### ### ### ###

   $indir  = "treedec";
   $outdir = "treedec";

   foreach $arg (@ARGV)
   {
      if ($arg =~ m%\Aindir=(\S+)\Z%)
      {  $indir = $1;
      }
      elsif ($arg =~ m%\Aoutdir=(\S+)\Z%)
      {  $outdir = $1;
      }
      elsif ($arg =~ m%\Adir=(\S+)\Z%)
      {  $outdir = $1;
         $indir  = $1;
      }
      else
      {  print ERRFILE "Mystery parameter: $arg";
	 $error_count++;
      }
   }

   $page_out   = ">$outdir/pages.dat";
   $link_out   = ">$outdir/links.dat";
   $url2nn_out = ">$outdir/url2nn.dat";
   $msg_out    = ">$outdir/convert-ll-msg.txt";

   open (ERRFILE, $msg_out) || die "Cannot open ERRFILE = $msg_out \n";
   $error_count = 0;

   # poor man's .mime_types ... 
   %file_type = 
                 ('html'  => 'HTML',
                  'htm'   => 'HTML',
                  'shtml' => 'HTML',
                  'shtm'  => 'HTML',
                  'css'   => 'HTML',

                  'prt'   => 'TEXT',
                  'txt'   => 'TEXT',
                  'asc'   => 'TEXT',

                  'ps'    => 'PRINT',
                  'pdf'   => 'PRINT',
                  'doc'   => 'PRINT',
                  'ppt'   => 'PRINT',

                  'vrml'  => '3D',
                  'wrl'   => '3D',

                  'jpg'   => 'IMAGE',
                  'jpeg'  => 'IMAGE',
                  'gif'   => 'IMAGE',
                  'tiff'  => 'IMAGE',
                  'png'   => 'IMAGE',

                  'mpg'   => 'VIDEO',
                  'mpeg'  => 'VIDEO',
                  'movie' => 'VIDEO',
                  'mov'   => 'VIDEO',
                  'mv'    => 'VIDEO',
                  'qt'    => 'VIDEO',
                  'rm'    => 'VIDEO',
                  'avi'   => 'VIDEO',
                  'smi'   => 'VIDEO',
                  'smil'  => 'VIDEO',

                  'cgi'   => 'EXEC',
                  'perl'  => 'EXEC',
                  'tcl'   => 'EXEC',
                  'tk'    => 'EXEC',
                  'java'  => 'EXEC',
                  'jav'   => 'EXEC',
                  'js'    => 'EXEC',
                  'class' => 'EXEC',
                  'perl'  => 'EXEC',
                  'pl'    => 'EXEC',
                  'exe'   => 'EXEC',

                  'mp3'   => 'AUDIO',
                  'au'    => 'AUDIO',
                  'wav'   => 'AUDIO' );

   ## Also, WEBSITE, MAIL, UNKNOWN, SUFFIX, and DIR are applied ad hoc;
   ## For now, we are relying on the fact that the first character 
   ## of each label is unique - can use it as abbreviation: WMUSDHTP3IVEA

### First pass - find all pages and assign nicknames

   open (LLFILE, "<$indir/fileF.txt")
      || die "Cannot open LLFILE in $indir\n";

   open  (PGFILE, $page_out) || die "Cannot open PGFILE = $page_out \n";
   print PGFILE "HOST=dummy\n";

   findline ("#-----------------") || die "Cannot find sep-1. \n";
   findline ("#-----------------") || die "Cannot find sep-2. \n";

   # collect all unique URLs
   $nick_ct = 0;
   $rec_ct = 8;

   while ($ll_line = <LLFILE>)
   {
      $rec_ct++;
      if ( ! ($rectype = parse_ll($ll_line)))
      {
         print ERRFILE "Strange line #$rec_ct: $ll_line\n";
	 $error_count++;
	 next;
      }

      $rectype =~ /[ftk]/            || next;
      if ($url_nn{$ll_url})
      {  # already got it
         next;
      }

      #OK, got new URL
      $ll_name = $ll_url;
      $ll_type = 'UNKNOWN';

      if ($ll_url =~ m%\Amailto:(\S+)@(\S+)\Z%)
      {
         $ll_name = "$2<$1";
         $ll_type = 'MAIL';
      }
      elsif ($ll_url =~ m%\w+://[\-\w\.]+\Z%)
      {
         $ll_type = 'WEBSITE';
      }
      elsif ($ll_url =~ m%(\S+)\.([a-zA-Z]{1,7})~?\Z%)
      {  # apparently found a file suffix
         $pref = $1;
         $suff = $2;
         $suff =~ tr/A-Z/a-z/;

         # found a known file suffix
         if ($file_type{$suff})
         {
            $ll_name = $pref;
            $ll_type = $file_type{$suff};
         }
         else
         {
            $ll_type = 'SUFFIX';
         }
      }
      elsif ($ll_url =~ m%/\Z%)
      {  # trailing slash means DIR - leave name as is
         $ll_type = 'DIR';
      }

      # gen unique nickname and make entries in url_nn and nn_url tables
      $namlen = length($ll_name);

      $nnlen = ($namlen < 10 ? $namlen : 10);
      # take last 10 or so chars and work your way back
      for ($nn_st = $namlen - $nnlen; $nn_st >= 0;  $nn_st--)
      {
         $nn = substr($ll_name, $nn_st, $nnlen);
         if ($nn_url{$nn})
         {} # this nickname already exists - keep trying
         else
         {  # OK, found a unique/new nickname
            last;
         }
      }

      if ($nn_st < 0)  #OK, try for meaningless, albeit unique nickname
      {
         for ($nn_st = $nick_ct; $nn_st < 1000000; $nn_st++)
         {
            $nn = 'N' . $nn_st;
            if ($nn_url{$nn})
            {} # this nickname already exists - keep trying
            else
            {  # OK, found a unique/new nickname
               last;
            }
         }
         $nn_st < 1000000
         || die "Cannot find unique nickname for $ll_url, rec#$rec_ct\n";

         $nick_ct = $nn_st;
         print "Assigned dumb nickname $nn to $ll_url.\n"
      }

      $url_nn{$ll_url} = $nn;
      $nn_url{$nn} = "$ll_url oftype $ll_type";
      ## print "Adding: $nn : $ll_url oftype $ll_type\n";
   }

   ## now gotta scan tables for equivalent URLS:
   ## xxx/a == xxx/a/ == xxx/a/index.htm == xxx/a/index.html
   ## ignore possibility of "index.HTML" for now...
   ## If you find equivalents, point shorter URL to nickname of longer.

   @target_suffix = ("index.html", "index.htm", "");

   foreach $nn (keys %nn_url)
   {
      $nnval = $nn_url{$nn};
      $nnval =~ m%\A(\S+) oftype (\S+)\Z%
         || die "$nn / $nnval did not match correct format.\n";
      $this_url  = $1;
      $this_type = $2;

      if ($this_type eq 'UNKNOWN')
      {  $slash = "/";
         $tries = 3;
      }
      elsif ($this_type eq 'DIR')
      {  $slash = "";
         $tries = 2;
      }
      else
      {  $tries = 0;
      }

      $kept_own_name = 1;
      for ( $ix=0; $ix < $tries && $kept_own_name; $ix++)
      {
         $target = $this_url . $slash . $target_suffix[$ix];
         $kept_own_name = subst_url ($this_url, $target);
      }

      if ($kept_own_name)
      {
         $short_type = substr($this_type,0,1);
         print PGFILE "$url_nn{$this_url} $short_type $this_url\n";
      }
   }

   close (PGFILE);

   ## write out url->nn table for later use in matching up webvip URLs

   open (U2FILE, $url2nn_out) || die "Cannot open U2FILE = $url2nn_out \n";
   print U2FILE "HOST=dummy\n";

   foreach $url (keys %url_nn)
   {
      print U2FILE "$url --->>> $url_nn{$url}\n";
   }
   close (U2FILE);

### Second pass - find links among unique pages

   print "Got URLs - now make links.\n";

   seek (LLFILE, 0, 0);
   open (LNFILE, $link_out) || die "Cannot open LNFILE = $link_out \n";

   findline ("Linklint version: ") || die "Confusing file header/1.\n";
   findline ("#-----------------") || die "Confusing file header/2.\n";
   findline ("#-----------------") || die "Confusing file header/3.\n";

   $st = 'expect_from';
   $rec_ct = 8;

   while ($ll_line = <LLFILE>)
   {
      $rec_ct++;

      if ( ! ($rectype = parse_ll($ll_line)))
      {
         print ERRFILE "Strange line (2nd pass, rec #$rec_ct ): $ll_line";
	 $error_count++;
	 next;
      }

      if ($rectype =~ /[ft]/)
      {  # get nn and true_url from ll_url
         if ($nn = $url_nn{$ll_url})
         {
            defined ($nn_url{$nn})
              || die "Mysterious URL/nn in rec #$rec_ct not found in table: $ll_url/$nn\n";
         }
         else
         {  die "Mysterious URL in rec #$rec_ct not found in table: $ll_url\n";
         }
      }

      if ($st eq 'expect_from')
      {
         if ($rectype eq 'f')
         {  $st = 'got_from';
            $from_nn = $nn;
         }
         else
         {  die "Expecting from-link, but got rec #$rec_ct: $ll_line\n";
         }
      }
      elsif ($st eq 'got_from')
      {
         if ($rectype eq 'c')
         {  $st = 'expect_to';
         }
         elsif ($rectype eq 'f')
         {  $st = 'got_from';
            $from_nn = $nn;
         }
         elsif ($rectype =~ m%[dk]%)
         {  $st = 'got_from';
         }
         else
         {  die "Expecting from-link or contains, but got rec #$rec_ct: $ll_line\n";
         }
      }
      elsif ($st eq 'expect_to')
      {
         if ($rectype eq 'b')
         {  $st = expect_from;
         }
         elsif ($rectype eq 't')
         {
            print LNFILE "$from_nn $nn\n";
         }
         else
         {  die "Expecting to-link or blank but got rec #$rec_ct: $ll_line\n";
         }
      }
      else
      {  die "Weird state = $st\n";
      }
   }

   close (LLFILE);
   close (LNFILE);
   close (ERRFILE);
   print "Convert-ll done. #errors detected = $error_count\n\n";

